import asyncio

from asyevent.callback import Callback
from asyevent.event import Event
from asyevent.command import Command

from asyevent.exceptions import CommandNotFound, EventNotFound, CommandAlreadyRegistered, EventAlreadyRegistered

from typing import Callable, Optional, Union, List
from copy import copy


class EventManager:
    """
    The manager stores events and commands, it
    provides an error handler for them.

    An event or command cannot be part of multiple event managers, because an event or command
    need a `event_manager` variable, to be executed directly with the error handler provided.

    """
    def __init__(self):
        """
        Initialises an event manager.
        """
        self._events: List[Event] = []
        self._commands: List[Command] = []

        self.loop = asyncio.get_event_loop()

        # For all events and commands that defines `handle_errors` set to True, their
        # exceptions are handled in this event.
        # Passed parameters are : the exception, the event, the callback, args, **kwargs.
        self.error_handler = self.create_event('<error_handler>', handle_errors=False)

    @property
    def events(self) -> List[Event]:
        """
        Property that makes registered events immutable from outside.

        :return: A copy of registered events.
        """
        return copy(self._events)

    @property
    def commands(self) -> List[Command]:
        """
        Property that makes registered commands immutable from outside.

        :return: A copy of registered commands.
        """
        return copy(self._commands)

    def as_command(
            self, name: str = None, handle_errors: bool = True,
            multiple_callbacks: bool = True, priority: int = 1, **options
    ) -> Callable[[Callable], Callback]:
        """
        A decorator which registers a coroutine as a callback of this command.

        :param priority: When an event is raised, associated callbacks will be invoked by ascending priority order.
        :param name: The command name. Coroutine's name by default.
        :param handle_errors: Determines if exceptions are handle into the error handler event.
        :param multiple_callbacks: Does the command allow multiple callbacks.

        :return: A callback.
        """

        def decorator(coroutine: Union[Callable, Callback]) -> Callback:
            command = self.create_command(
                coroutine=coroutine,
                name=name,
                handle_errors=handle_errors,
                multiple_callbacks=multiple_callbacks,
                is_classmethod=False,
                priority=priority,
                **options
            )

            return command.initial_callback

        return decorator

    def as_class_command(
            self, name: str = None, handle_errors: bool = True,
            multiple_callbacks: bool = True, priority: int = 1, **options
    ) -> Callable[[Callable], Callback]:
        """
        A decorator which registers a classmethod coroutine as a callback of this command.

        :param priority: When an event is raised, associated callbacks will be invoked by ascending priority order.
        :param name: The command name. Coroutine name by default.
        :param handle_errors: Determines if exceptions are handle into the error handler event.
        :param multiple_callbacks: Does the command allow multiple callbacks.

        :return: A callback.
        """

        def decorator(coroutine: Union[Callable, Callback]) -> Callback:
            command = self.create_command(
                coroutine=coroutine,
                name=name,
                handle_errors=handle_errors,
                multiple_callbacks=multiple_callbacks,
                is_classmethod=True,
                priority=priority,
                **options
            )

            return command.initial_callback

        return decorator

    def create_command(
            self, coroutine: Union[Callable, Callback], name: str = None, handle_errors: bool = True,
            multiple_callbacks: bool = True, is_classmethod: bool = False, priority: int = 1, **options
    ) -> Command:
        """
        Registers a coroutine as a `command` event.
        The purpose of using this instead of `Event.as_callback` is dynamic event names.

        :param multiple_callbacks: Does the command allow multiple callbacks associated to.
        :param coroutine: Coroutine which will be called when the command event is raised.
        :param name: The command name. Coroutine name by default.
        :param handle_errors: Determines if exceptions are handle into the error handler event.
        :param is_classmethod: Defines if the callback must be invoked with a `self` parameter.
        :param priority: When an event is raised, associated callbacks will be invoked by ascending priority order.
        :param options: Callback options

        :return: A command.
        """
        command = Command(
            coroutine=coroutine,
            event_manager=self,
            name=name,
            handle_errors=handle_errors,
            multiple_callbacks=multiple_callbacks,
            is_classmethod=is_classmethod,
            priority=priority,
            **options
        )
        self._commands.append(command)

        return command

    def create_event(self, name: str, handle_errors: bool = True, multiple_callbacks: bool = True) -> Event:
        """
        Creates an event associated to the manager.

        :param multiple_callbacks: Does the command allow multiple callbacks associated to.
        :param name: The event name.
        :param handle_errors: Determines if exceptions are handle into the error handler event.

        :return: The event.
        """
        event = Event(
            name=name,
            event_manager=self,
            handle_errors=handle_errors,
            multiple_callbacks=multiple_callbacks
        )
        self._events.append(event)

        return event

    def get_command(self, name: str, case_sensitive: bool = True) -> Optional[Command]:
        """
        Get a command by its name. The command must be associate to this manager.

        :param name: The command name.
        :param case_sensitive: Is case sensitive.

        :return: A command if found.
        """
        return {
            cmd.command_name.lower() if not case_sensitive else cmd.command_name: cmd for cmd in self._commands
        }.get(name.lower() if not case_sensitive else name)

    def get_event(self, name: str, case_sensitive: bool = True) -> Optional[Event]:
        """
        Get an event by its name. The event must be associate to this manager.

        :param name: The event name.
        :param case_sensitive: Is case sensitive.

        :return: An event if found.
        """
        return {
            event.event_name.lower() if not case_sensitive else event.event_name: event for event in self._events
        }.get(name.lower() if not case_sensitive else name)

    def replace_command_name(self, name: str, new_name: str):
        """
        Replace a command name.

        :param name: The name of the command to replace.
        :param new_name: The new name.

        :raise: CommandNotFound: If no command was found.
        """
        command = self.get_command(name)

        if not command:
            raise CommandNotFound(name=name)

        command.command_name = new_name

    def add_event(self, event: Event):
        """
        Add an event to this event manager, and remove it from the previous one.

        :param event: The event to add.

        :raise: EventAlreadyRegistered: If the event is already registered.
        """
        if event in self._events:
            raise EventAlreadyRegistered(event=event)

        event.event_manager.remove_event(event)
        self._events.append(event)
        event.event_manager = self

    def remove_event(self, event: Event):
        """
        Remove an event from this event manager.

        :param event: The event to remove.
        """
        self._events.remove(event)
        event.event_manager = None

    def remove_event_by_name(self, name: str):
        """
        Remove an event by its name from this event manager.

        :param name: The name of the event to remove.

        :raise: EventNotFound: If no event was found.
        """
        event = self.get_event(name)

        if not event:
            raise EventNotFound(name=name)

        self.remove_event(event)

    def add_command(self, command: Command):
        """
        Add a command to this event manager, and remove it from the previous one.

        :param command: The command to add.

        :raise: CommandAlreadyRegistered: If the command is already registered.
        """
        if command in self._commands:
            raise CommandAlreadyRegistered(command=command)

        command.event_manager.remove_command(command)
        self._commands.append(command)
        command.event_manager = self

    def remove_command(self, command: Command):
        """
        Remove a command from this event manager.

        :param command: The command to remove.
        """
        self._commands.remove(command)
        command.event_manager = None

    def remove_command_by_name(self, name: str):
        """
        Remove a command by its name from this event manager.

        :param name: The name of the command to remove.

        :raise: CommandNotFound: If no command was found.
        """
        command = self.get_command(name)

        if not command:
            raise CommandNotFound(name=name)

        self.remove_command(command)

    async def invoke_command(self, name: str, *args, case_sensitive: bool = True, **kwargs):
        """
        Invokes a command by its name.

        :param name: The command name.
        :param args: The command parameters.
        :param case_sensitive: Is case sensitive.
        :param kwargs: The command keyword parameters.

        :raise CommandNotFound: If no command was found.:
        """
        command = self.get_command(name, case_sensitive=case_sensitive)

        if not command:
            raise CommandNotFound(name)

        await command.raise_event(*args, **kwargs)

    async def raise_event(self, name: str, *args, case_sensitive: bool = True, **kwargs):
        """
        Raises an event by its name.

        :param name: The event name.
        :param case_sensitive: Is case sensitive.
        :param args: The event parameters.
        :param kwargs: The event keyword parameters.

        :raise EventNotFound: If no event was found.:
        """
        event = self.get_event(name, case_sensitive=case_sensitive)

        if not event:
            raise EventNotFound(name)

        await event.raise_event(*args, **kwargs)



"""
All specific package's exceptions raised.
"""


class EventSystemException(Exception):
    """
    Base exception.
    """
    pass


class EventException(EventSystemException):
    """
    Base event exception.
    """
    def __init__(self, message: str, event=None):
        self.event = event
        super().__init__(message)


class CommandException(EventSystemException):
    """
    Base command exception.
    """
    def __init__(self, message: str, command=None):
        self.command = command
        super().__init__(message)


class EventNotFound(EventException):
    """
    Raise when a unknown event is invoked.
    """
    def __init__(self, name: str):
        self.name = name
        super().__init__(f'Event {self.name!r} cannot be found.')


class EventAlreadyExists(EventException):
    """
    Raise when an event that is already defined by its event_name is created.
    """
    def __init__(self, event):
        super().__init__(f'Event {event.event_name!r} already exists.', event)


class EventAlreadyRegistered(EventException):
    """
    Raise when an event is add to en event_manager that already contains the event.
    """
    def __init__(self, event):
        super().__init__(f'Event {event.event_name!r} is already registered.', event)


class CommandNotFound(CommandException):
    """
    Raise when a unknown command is invoked.
    """
    def __init__(self, name: str):
        self.name = name
        super().__init__(f'Command {self.name!r} cannot be found.')


class CommandAlreadyExists(CommandException):
    """
    Raise when an event that is already defined by its event_name is created.
    """
    def __init__(self, command):
        super().__init__(f'Command {command.command_name!r} already exists.', command)


class CommandAlreadyRegistered(CommandException):
    """
    Raise when a command is add to en event_manager that already contains the command.
    """
    def __init__(self, command):
        super().__init__(f'Command {command.command_name!r} is already registered.', command)



from __future__ import annotations

import asyncio
import time

from asyevent.callback import Callback
from asyevent.exceptions import EventAlreadyExists

from typing import Callable, Union, List, Dict


class Event:
    """
    Do not manually initialise this class. Use instead `EventManager.create_event()`.

    Events contain `Callback` objects. They can be raised for invoke all callbacks stored.
    """
    def __init__(self, name: str, event_manager, handle_errors: bool = True, multiple_callbacks: bool = True):
        """
        Initialises an event.

        :param name: The event name. Must be unique in the event manager.
        :param event_manager: An event manager.
        :param handle_errors: Do the errors handled into the `event_manager` error handler.
        :param multiple_callbacks: Does the command allow multiple callbacks associated to.

        :raise EventAlreadyExists: If the event name is not unique in the event_manager.
        """
        self.event_name = name

        if event_manager.get_event(self.event_name, case_sensitive=False):
            raise EventAlreadyExists(event_manager.get_event(self.event_name))

        self.handle_errors = handle_errors
        self.event_manager = event_manager
        self.multiple_callbacks = multiple_callbacks

        self._before = None
        self._after = None

        self._loop = asyncio.get_event_loop()
        self._tasks: List[asyncio.Task] = []
        self._callbacks: Dict[int, List[Callback]] = {}

    def __iadd__(self, callback: Callback):
        # default priority, use `.add_callback()` to custom it.
        self.add_callback(callback)

        return self

    def __isub__(self, callback: Callback):
        self.remove_callback(callback)

        return self

    async def __call__(self, *args, **kwargs):
        await self.raise_event(*args, **kwargs)

    @property
    def before(self) -> Event:
        """
        Returns an event that is raised before callbacks.
        The current parameters will be passed to the callback.
        """
        if not self._before:
            self._before = self.event_manager.create_event(
                f'<before:{self.event_name}>',
                handle_errors=self.handle_errors
            )

        return self._before

    @property
    def after(self) -> Event:
        """
        Returns an event that is raised after the callbacks.
        Their execution duration in seconds as
        well as the current parameters, will be passed to the callback.
        """
        if not self._after:
            self._after = self.event_manager.create_event(
                f'<after:{self.event_name}>',
                handle_errors=self.handle_errors
            )

        return self._after

    @property
    def callbacks(self) -> List[Callback]:
        """
        Sorts callbacks by layer level and return them.

        :return: :class: `List[Callback]`
        """

        result: List[Callback] = []

        for _, callbacks in reversed(sorted(self._callbacks.items())):
            result.extend(callbacks)

        return result

    def as_callback(self, priority: int = 1, **options) -> Callable[[Callable], Callback]:
        """
        A decorator which registers a coroutine as a callback of this event.
        All event's callbacks are invoked when `raise_event` is called.

        :return: `Callback` object.
        """

        def decorator(coroutine: Union[Callable, Callback]) -> Callback:
            return self.create_callback(
                coroutine=coroutine,
                is_classmethod=False,
                priority=priority,
                **options
            )

        return decorator

    def as_class_callback(self, priority: int = 1, **options) -> Callable[[Callable], Callback]:
        """
        A decorator which registers a  classmethod coroutine as a callback of this event.
        All event's callbacks are invoked when `raise_event` is called.

        :return: `Callback` object.
        """

        def decorator(coroutine: Union[Callable, Callback]) -> Callback:
            return self.create_callback(
                coroutine=coroutine,
                is_classmethod=True,
                priority=priority,
                **options
            )

        return decorator

    def create_callback(
            self, coroutine: Union[Callable, Callback], is_classmethod: bool = False, priority: int = 1, **options
    ) -> Callback:
        """
        Create a callback and add it to this event.

        :param coroutine: The coroutine that will be
        :param is_classmethod: Does the callback need a `self` parameter.
        :param priority: When the event is raised, callbacks are invoked in priority ascending order.
        :param options: Callback options.

        :return: The callback created.
        """
        callback = Callback(
            coroutine=coroutine,
            is_classmethod=is_classmethod,
            **options
        )

        self.add_callback(
            callback=callback,
            priority=priority
        )

        return callback

    def add_callback(self, callback: Callback, priority: int = 1):
        """
        Registers a callback to this event.

        :param callback: The callback to register.
        :param priority: When the event is raised, callbacks are invoked in priority ascending order.

        :raise ValueError: If there is multiple callbacks and `multiple_callbacks` is `False`.:
        :raise ValueError: If the callback is already registered.:
        """
        if not self.multiple_callbacks and self.callbacks:
            raise ValueError(f'Cannot add multiple callbacks on event {self.event_name!r}.')

        if callback in self.callbacks:
            raise ValueError(f'Callback {callback.__name__!r} is already registered.')

        if self._callbacks.get(priority) is None:
            self._callbacks[priority] = []

        self._callbacks[priority].append(callback)

    def remove_callback(self, callback: Callback):
        """
        Removes a callback from this event.

        :param callback: The callback to remove.

        :raise ValueError: If the callbacks is not registered in this event.:
        """
        new = {p: [c_ for c_ in c if c_ != callback] for p, c in self._callbacks.items()}

        if new == self._callbacks:
            raise ValueError(f'Callback {callback.__name__!r} is not registered.')

        self._callbacks = new

    async def raise_event(self, *args, **kwargs):
        """
        Calls all registered callbacks. If `handle_errors` is True, `error_handler` event will be raised.
        Parameters passed are : :class: `Exception` the error, :class: `Event` the event where the error occurred,
        *args and **kwargs passed into the initial event.

        :param args: Invocation parameters.
        :param kwargs: Invocation keyword parameters.
        """
        if self._before and self.before.callbacks:
            await self.before.raise_event(*args, **kwargs)

        start_time = time.time()
        for callback in self.callbacks:
            self._tasks.append(self._loop.create_task(callback.invoke(
                *args,
                **kwargs,
                _event=self,
                _handler=self.event_manager.error_handler if self.handle_errors else None
            )))

        await asyncio.gather(*self._tasks)

        if self._after and self.after.callbacks:
            await self.after.raise_event(time.time() - start_time, *args, **kwargs)

    def cancel(self):
        """
        Cancel all callbacks that run.
        """
        for task in self._tasks:
            task.cancel()



from __future__ import annotations

import asyncio
import inspect

from typing import Callable, Union, Tuple


class Callback:
    """
    Callable objet, wrap the coroutine behavior with options and errors handler.
    """
    def __init__(
            self, coroutine: Union[Callable, Callback], is_classmethod: bool = False, **options
    ):
        """
        Initialises a callback with a coroutine and options.

        :param coroutine: The coroutine that will be executed on invoke.
        :param is_classmethod: Does the coroutine need a `self` parameter.
        :param options: Options describing how the coroutine behave.

        :raise TypeError: If the `coroutine` parameter is not a coroutine.:
        :raise ValueError: If the callback is a classmethod but miss the `self` parameter.:
        """
        if isinstance(coroutine, Callback):
            # copy old callback's attributes.
            self.__dict__ = coroutine.__dict__

            return

        self._coroutine = coroutine
        self.__name__ = self._coroutine.__name__

        self.is_classmethod = is_classmethod

        if not inspect.iscoroutinefunction(coroutine):
            raise TypeError(f'Callback function {self.__name__!r} must be a _coroutine.')

        if self.is_classmethod and self.signature.parameters.get('self') is None:
            raise ValueError(f'Instance parameter \'self\' is missing on {self.__name__!r} callback.')

        self.is_running = False
        self.wrapper = None

        # is the coroutine active
        self.is_active: bool = options.get('is_active', True)

        # the delay in seconds before the coroutine is called.
        # it does not impact the event raising process.
        self.start_delay: float = options.get('start_delay', 0)

        # Even if an handler is pass in `.invoke()` method,
        # the exceptions will be raised if it's set to `True`.
        self.refuse_handling = options.get('refuse_handling', False)

        # loop settings
        # the coroutine will be called `repeat_times` times.
        self.repeat_times: int = options.get('repeat_times', options.get('loop', 1))
        # the delay in seconds between every loop iteration of given by `repeat_times`.
        self.loop_delay: float = options.get('loop_delay', 0)
        # when a exception occur, if it's handled, defines if the loop iterations should continue.
        self.continue_on_error: bool = options.get('continue_on_error', False)

    @property
    def signature(self) -> inspect.Signature:
        """
        The coroutine signature.

        :return: The coroutine signature.
        """
        return inspect.signature(self._coroutine)

    async def invoke(self, *args, _event=None, _handler=None, **kwargs):
        """
        Invokes the coroutine.

        :param _event: If an error is handled, the event will be passed as a parameter.
        :param _handler: If an error occur, the `_handler` event will be raised.
        :param args: Coroutine parameters.
        :param kwargs: Coroutine keyword parameters.

        :raise Exception: If an exception cannot be handle, it will be raised.:
        """
        if self.is_active:
            self.is_running = True
            await asyncio.sleep(self.start_delay)

            async def iterate():
                # stop iteration if an error is returned.
                if isinstance(await self._invoke_once(
                        *args, **kwargs, _event=_event, _handler=_handler,
                ), Exception):
                    return

            if self.repeat_times <= 0:
                while True:
                    await iterate()

            else:
                for _ in range(self.repeat_times):
                    await iterate()

            self.is_running = False

    async def _invoke_once(self, *args, _event=None, _handler=None, **kwargs) -> Exception:
        parameters = self._parse_arguments(*args, **kwargs)

        try:
            await self._coroutine(*parameters[0], **parameters[1])

        except Exception as e:
            if _handler and not self.refuse_handling:
                asyncio.create_task(_handler.raise_event(e, _event, self, *args, **kwargs))

                if not self.continue_on_error:
                    return e

            else:
                raise e

        if self.repeat_times > 1:
            await asyncio.sleep(self.loop_delay)

    def _parse_arguments(self, *args, **kwargs) -> Tuple[list, dict]:
        if self.is_classmethod:
            if self.wrapper is None:
                raise ValueError(
                    f'Missing wrapper instance on classmethod callback {self.__name__!r}. Commun causes are : '
                    f'\n * Missing to to call EventWrapper.init_callbacks().'
                    f'\n * Declaring callback as class callbacks.'
                    f'\n * Overriding this callback method.'
                )

            args = [self.wrapper] + list(args)

        return args, kwargs

    def enable(self):
        """
        Enables the callback.
        """
        self.is_active = True

    def disable(self):
        """
        Disables the callback.
        """
        self.is_active = False

    def loop(self, times: int, delay: float = 0):
        """
        Shortcut for loop settings.

        :param times: How many times the _coroutine iterate on invocation.
        :param delay: Delay between _coroutine iterations.
        """
        self.repeat_times = times
        self.loop_delay = delay



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="asyevent",
    version="0.1.3",
    description="An implementation of events and asynchronous callbacks using decorators.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/HerveZu/asyevent",
    author="Zucchinetti Hervé",
    author_email="herve.zucchinetti@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["asyevent"],
    include_package_data=True,
    install_requires=["asyncio"],
)
